#!/usr/bin/env python

"""

Utility script used to sync a single project from pkgdb to pagure with the
added information provided.

This script relies on what having the main admin or group already set on
pagure.

"""


import argparse
import collections
import logging
import os

import requests
import pygit2
from sqlalchemy.exc import SQLAlchemyError

_log = logging.getLogger(__name__)

if 'PAGURE_CONFIG' not in os.environ \
        and os.path.exists('/etc/pagure/pagure.cfg'):
    print('Using configuration file `/etc/pagure/pagure.cfg`')
    os.environ['PAGURE_CONFIG'] = '/etc/pagure/pagure.cfg'

import pagure
import pagure.lib
import pagure.lib.model


PKGDB_URL = 'https://admin.fedoraproject.org/pkgdb/api/'
BLACKLIST = ['@provenpackager', 'group::provenpackager']


def get_pkg_info(namespace, pkg):
    """ Retrieve a dictionary giving the maintainers of each package
    in pkgdb.
    """
    _log.debug('Querying pkgdb')
    req = requests.get(PKGDB_URL + 'package/%s/%s' % (namespace, pkg))
    _log.debug('Pkgdb data retrieved, getting ACLs')
    acls = collections.defaultdict(set)
    data = req.json()['packages']
    for pkg in data:
        for acl in pkg['acls']:
            if acl['status'] == 'Approved':
                acls[acl['acl']].add(acl['fas_name'])

    return acls


def main(namespace, project, poc):
    """
    Logic:
    - Retrieve the pkgdb information about this project
    - Create the project in pagure with the provided main admin and the
      rest of the information from pkgdb.

    """

    _log.debug('namespace:     %s', namespace)
    _log.debug('project:       %s', project)
    _log.debug('poc:           %s', poc)

    pkg_info = get_pkg_info(namespace, project)

    fullname = '%s/%s' % (namespace, project)
    try:
        if fullname in pagure.APP.config['BLACKLISTED_PROJECTS']:
            raise pagure.exceptions.RepoExistsException(
                'No project "%s" are allowed to be created due to '
                'potential conflicts in URLs with pagure itself' % project
            )

        user_obj = pagure.lib.get_user(pagure.SESSION, poc)
        allowed_prefix = pagure.APP.config[
            'ALLOWED_PREFIX'] + [grp for grp in user_obj.groups]

        if namespace not in allowed_prefix:
            raise pagure.exceptions.PagureException(
                'The prefix of your project must be in the list of '
                'allowed prefixes set by the admins of this pagure '
                'instance, or the name of a group of which you are a '
                'member.'
            )

        project = pagure.lib.model.Project(
            name=project,
            namespace=namespace,
            description='The %s %s' % (project, namespace),
            url=None,
            avatar_email=None,
            user_id=user_obj.id,
            parent_id=None,
            hook_token=pagure.lib.login.id_generator(40)
        )
        pagure.SESSION.add(project)
        # Make sure we won't have SQLAlchemy error before we create
        # the repo
        pagure.SESSION.flush()

        _log.info('Set commit and admin access')
        for acl in pkg_info:
            if acl not in ['commit', 'approveacls']:
                continue

            access = 'commit'
            if acl == 'approveacls':
                access = 'admin'

            for maintainer in pkg_info[acl]:
                maintainer = maintainer.strip()
                if not maintainer:
                    continue
                _log.info('  Processing: %s', maintainer)
                if maintainer in project.contributors['commit']:
                    _log.debug('    - %s has already commit' % maintainer)
                    continue
                try:
                    if maintainer.startswith(('@', 'group::')):
                        if maintainer in BLACKLIST:
                            continue
                        if maintainer.startswith('@'):
                            name = maintainer[1:]
                        else:
                            name = maintainer[7:]

                        out = pagure.lib.add_group_to_project(
                            session=pagure.SESSION,
                            project=project,
                            new_group=name,
                            user='releng',
                            access=access,
                            is_admin=True)
                        _log.debug('    - %s', out)
                    else:
                        out = pagure.lib.add_user_to_project(
                            session=pagure.SESSION,
                            project=project,
                            new_user=maintainer,
                            user='releng',
                            access=access)
                        _log.debug('    - %s', out)
                    pagure.SESSION.commit()
                except pagure.exceptions.PagureException as err:
                    print('ERROR with project %s' % fullname)
                    print(err)

        _log.info('Set watch ticket flags')
        # Set-up the watch-ticket flags
        for maintainer in pkg_info['watchbugzilla']:
            maintainer = maintainer.strip()
            if not maintainer:
                continue
            _log.info('  - %s' % maintainer)

            if maintainer.startswith(('@', 'group::')):
                _log.debug('    Skipping group')
                continue

            watch = '1'
            if maintainer in pkg_info['watchcommits']:
                watch = '3'

            try:
                out = pagure.lib.update_watch_status(
                    session=pagure.SESSION,
                    project=project,
                    user=maintainer,
                    watch=watch
                )
                _log.info('    - %s', out)
            except pagure.exceptions.PagureException as err:
                print('ERROR with project %s' % fullname)
                print(err)

        _log.info('Set watch commit flags')
        # Set-up the watch-commits flags
        for maintainer in pkg_info['watchcommits']:
            maintainer = maintainer.strip()
            if not maintainer:
                continue
            _log.debug('  - %s' % maintainer)

            if maintainer.startswith(('@', 'group::')):
                _log.debug('    Skipping group')
                continue

            watch = '2'
            if maintainer in pkg_info['watchbugzilla']:
                watch = '3'

            try:
                out = pagure.lib.update_watch_status(
                    session=pagure.SESSION,
                    project=project,
                    user=maintainer,
                    watch=watch
                )
                _log.info('    - %s', out)
            except pagure.exceptions.PagureException as err:
                print('ERROR with project %s' % fullname)
                print(err)

        gitfolder = pagure.APP.config['GIT_FOLDER']
        requestfolder = pagure.APP.config['REQUESTS_FOLDER']

        gitrepo = os.path.join(gitfolder, '%s.git' % fullname)
        http_clone_file = os.path.join(gitrepo, 'git-daemon-export-ok')
        if not os.path.exists(http_clone_file):
            with open(http_clone_file, 'w'):
                pass

        requestrepo = os.path.join(requestfolder, project.path)
        if not os.path.exists(requestrepo):
            pygit2.init_repository(
                requestrepo, bare=True,
                mode=pygit2.C.GIT_REPOSITORY_INIT_SHARED_GROUP)
        else:
            _log.info('  - The requests repo "%s" already exists' % (
                project.path))
            pass

        pagure.SESSION.commit()
    except pagure.exceptions.PagureException as err:
        print('ERROR with project %s' % fullname)
        print(err)
    except SQLAlchemyError as err:  # pragma: no cover
        pagure.SESSION.rollback()
        print('ERROR (DB) with project %s' % project)
        print(err)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Script creating a single projects on pagure based on '
        'provided information and what is in pkgdb.'
    )
    parser.add_argument(
        'namespace', help='The namespace of the project')
    parser.add_argument(
        'project', help='The name of the project')
    parser.add_argument(
        '--poc', dest='poc', help='The main admin of the project')
    parser.add_argument(
        '--debug', dest='debug', action='store_true', default=False,
        help='Print the debugging output')

    args = parser.parse_args()

    main(
        namespace=args.namespace,
        project=args.project,
        poc=args.poc
    )
