from __future__ import print_function

import dist_git_auth

from mock import Mock, patch

import pagure.config
import pagure.exceptions
import pagure.lib.model

try:
    from pagure.lib import _get_project as get_project, add_user_to_project
except ImportError:
    # From pagure 5.2, code has been moved to pagure.lib.query
    from pagure.lib.query import (
        _get_project as get_project,
        add_user_to_project,
    )

import tests


class DistGitAuthTests(tests.Modeltests):
    """Test DistGitAuth ACLs with Fedora config."""

    maxDiff = None

    def setUp(self):
        """Set up the environment in which to run the tests."""
        super(DistGitAuthTests, self).setUp()

        pagure.config.config["ACL_DEBUG"] = True
        pagure.config.config["EXTERNAL_COMMITTER"] = {"relenggroup": {}}
        pagure.config.config.update(self.dga_config)

        self.dga = dist_git_auth.DistGitAuth()
        self.dga.info = Mock(side_effect=print)
        # We default to saying it's not a forced push
        dist_git_auth.is_forced_push = Mock(return_value=False)

        # Create an RCM user/group
        rcmuser = pagure.lib.model.User(
            user="releng",
            fullname="Release Engineering",
            token="aaabbbcd",
            default_email="rcm@local.local",
        )
        self.session.add(rcmuser)
        self.session.flush()
        rcmgroup = pagure.lib.model.PagureGroup(
            group_name="relenggroup",
            group_type="user",
            display_name="Releng group",
            user_id=rcmuser.id,
        )
        self.session.add(rcmgroup)
        self.session.flush()
        rcmuser.group_objs.append(rcmgroup)
        self.session.commit()

    def tearDown(self):
        """Tear down the environment in which the tests ran."""
        self.dga = None
        super(DistGitAuthTests, self).tearDown()

    def create_namespaced_project(
        self,
        namespace,
        name,
        is_fork=False,
        settings={},
        hook_token_postfix="",
    ):
        item = pagure.lib.model.Project(
            user_id=1,  # pingou
            name=name,
            is_fork=is_fork,
            parent_id=3 if is_fork else None,
            description="namespaced test project",
            hook_token="aaabbbeee" + hook_token_postfix,
            namespace=namespace,
            settings=settings,
        )
        item.close_status = [
            "Invalid",
            "Insufficient data",
            "Fixed",
            "Duplicate",
        ]
        self.session.add(item)
        self.session.commit()
        return get_project(self.session, name=name, namespace=namespace)

    def expect_info_msg(self, expect_msg):
        found = False
        for call in self.dga.info.call_args_list:
            args = call[0]
            msg = args[0]
            if msg == expect_msg:
                found = True
        if not found:
            raise AssertionError("Info message '%s' expected but not found" % expect_msg)


class DistGitAuthTestsGeneric(DistGitAuthTests):
    dga_config = {
        "BLACKLIST_RES": [],
        "RCM_BRANCHES": [],
        "BYPASS_PR_ONLY_GROUPS": ["relenggroup"],
    }

    def test_unused_repotype(self):
        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=None,
                username=None,
                refname=None,
                pull_request=None,
                repodir=None,
                repotype="tickets",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Repotype tickets not in use")

    def test_branch_deletion(self):
        dist_git_auth.is_forced_push.return_value = True

        project = self.create_namespaced_project("rpms", "test")

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username=None,
                refname=None,
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto="0000000000000000000000000000000000000000",
                is_internal=False,
            )
        )

        self.expect_info_msg("Branch deletion is not allowed")

    def test_forced_push(self):
        dist_git_auth.is_forced_push.return_value = True

        project = self.create_namespaced_project("rpms", "test")

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username=None,
                refname=None,
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Forced pushes are not allowed")

    def test_internal(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username=None,
                refname=None,
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=True,
            )
        )

        self.expect_info_msg("Internal push allowed")

    def test_deploykey(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="deploykey_foobar",
                refname=None,
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Deploy keys are disabled")

    def test_invalid_user(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertRaises(
            pagure.exceptions.PagureException,
            self.dga.check_acl,
            self.session,
            project=project,
            username="nosuchuser",
            refname=None,
            pull_request=None,
            repodir=None,
            repotype="main",
            revfrom=None,
            revto=None,
            is_internal=False,
        )

    def test_unprotected_committer(self):
        project = self.create_namespaced_project("unprotected", "test")

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer push")

    def test_unprotected_non_committer(self):
        project = self.create_namespaced_project("unprotected", "test")

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Fall-through deny")

    def test_unprotected_pr_required_pr(self):
        project = self.create_namespaced_project("unprotected", "test")
        self.dga.global_pr_only = True

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=True,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer push")

    def test_unprotected_pr_required_no_pr(self):
        project = self.create_namespaced_project("unprotected", "test")
        self.dga.global_pr_only = True

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("A pull request is required for this branch")

    def test_unprotected_pr_required_no_pr_cvsadmin(self):
        project = self.create_namespaced_project("unprotected", "test")
        self.dga.global_pr_only = True

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="releng",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer push")

    def test_unprotected_pr_required_requests(self):
        project = self.create_namespaced_project("unprotected", "test")
        self.dga.global_pr_only = True

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="requests",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer push")

    def test_unprotected_pr_required_repo_pr_only_no_pr(self):
        settings = {"pull_request_access_only": True}
        project = self.create_namespaced_project("unprotected", "test", settings=settings)

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("A pull request is required for this branch")

    def test_unprotected_pr_required_repo_pr_only(self):
        settings = {"pull_request_access_only": True}
        project = self.create_namespaced_project("unprotected", "test", settings=settings)
        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request="pull_request",
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer push")


class DistGitAuthTestsFedora(DistGitAuthTests):
    dga_config = {
        "PR_ONLY": False,
        "ACL_BLOCK_UNSPECIFIED": False,
        "BLACKLIST_RES": ["refs/heads/c[0-9]+.*"],
        "UNSPECIFIED_BLACKLIST_RES": ["refs/heads/f[0-9]+"],
        "RCM_GROUP": "relenggroup",
        "RCM_BRANCHES": ["refs/heads/f[0-9]+"],
        "ACL_PROTECTED_NAMESPACES": ["rpms", "modules", "container"],
        "PDC_URL": "invalid://",
    }

    def test_protected_blacklisted_ref(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/c7",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Ref refs/heads/c7 is blocked")

    def test_protected_rcm(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="releng",
                refname="refs/heads/f27",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("RCM push")

    @patch("dist_git_auth.requests")
    def test_protected_unsupported_branch(self, mock_requests):
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": [{"active": False}]}
        mock_requests.get.return_value = res
        project = self.create_namespaced_project("rpms", "test")

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/f26",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Branch refs/heads/f26 is unsupported. Cannot push to a disabled branch (maybe eol?).")

    @patch("dist_git_auth.requests")
    def test_protected_supported_branch_committer(self, mock_requests):
        project = self.create_namespaced_project("rpms", "test")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": [{"active": True}]}
        mock_requests.get.return_value = res

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/f27",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Branch refs/heads/f27 is supported")

    @patch("dist_git_auth.requests")
    def test_protected_supported_branch_non_committer(self, mock_requests):
        project = self.create_namespaced_project("rpms", "test")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": [{"active": True}]}
        mock_requests.get.return_value = res

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/f27",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Branch refs/heads/f27 is supported")

    @patch("dist_git_auth.requests")
    def test_protected_unspecified_branch_blacklisted(self, mock_requests):
        project = self.create_namespaced_project("rpms", "test")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/f28",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Unspecified ref refs/heads/f28 is blocked")

    @patch("dist_git_auth.requests")
    def test_protected_unspecified_branch_normal_committer(self, mock_requests):
        project = self.create_namespaced_project("rpms", "test")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Unspecified branch push")

    @patch("dist_git_auth.requests")
    def test_protected_unspecified_branch_normal_non_committer(self, mock_requests):
        project = self.create_namespaced_project("rpms", "test")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Unspecified branch push")


class DistGitAuthTestsFedoraCommitterAccess(DistGitAuthTests):
    dga_config = {
        "PR_ONLY": False,
        "ACL_BLOCK_UNSPECIFIED": False,
        "BLACKLIST_RES": ["refs/heads/c[0-9]+.*"],
        "UNSPECIFIED_BLACKLIST_RES": ["refs/heads/f[0-9]+"],
        "RCM_GROUP": "relenggroup",
        "RCM_BRANCHES": ["refs/heads/f[0-9]+"],
        "ACL_PROTECTED_NAMESPACES": ["rpms", "modules", "container"],
        "PDC_URL": "invalid://",
    }

    def setUp(self):
        super(DistGitAuthTestsFedoraCommitterAccess, self).setUp()

        project = self.create_namespaced_project("rpms", "test")

        msg = add_user_to_project(
            session=self.session,
            project=project,
            new_user="foo",
            user="pingou",
            access="collaborator",
            branches="epel*",
        )
        self.session.commit()
        self.assertEqual(msg, "User added")

    @patch("dist_git_auth.requests")
    def test_protected_unspecified_branch_collaborator_invalid_branch(
        self, mock_requests
    ):
        project = get_project(self.session, name="test", namespace="rpms")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/mywip",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer: False")
        self.expect_info_msg("Fall-through deny")

    @patch("dist_git_auth.requests")
    def test_protected_unspecified_branch_collaborator_valid_branch(self, mock_requests):
        project = get_project(self.session, name="test", namespace="rpms")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/epel8",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer: True")
        self.expect_info_msg("Committer push")


class DistGitAuthTestsCentOS(DistGitAuthTests):
    dga_config = {
        "PR_ONLY": False,
        "ACL_BLOCK_UNSPECIFIED": True,
        "BLACKLIST_RES": ["refs/heads/f[0-9]+.*"],
        "RCM_GROUP": "relenggroup",
        "RCM_BRANCHES": ["refs/heads/c[0-9]+"],
        "SUPPORTED_SIGS": ["sig-core"],
        "SIG_PREFIXES": ["refs/heads/c7"],
        "ACL_PROTECTED_NAMESPACES": ["rpms"],
    }

    def setUp(self):
        super(DistGitAuthTestsCentOS, self).setUp()

        # Create an RCM user/group
        arrfab = pagure.lib.model.User(
            user="arrfab",
            fullname="Fabian Arriton",
            token="aaabbbcd",
            default_email="arrfab@local.local",
        )
        self.session.add(arrfab)
        self.session.flush()
        sigcore = pagure.lib.model.PagureGroup(
            group_name="sig-core",
            group_type="user",
            display_name="Core SIG group",
            user_id=arrfab.id,
        )
        self.session.add(sigcore)
        self.session.flush()
        arrfab.group_objs.append(sigcore)
        self.session.commit()

    def test_protected_blacklisted_ref(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/f27",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Ref refs/heads/f27 is blocked")

    def test_protected_rcm(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="releng",
                refname="refs/heads/c7",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("RCM push")

    def test_protected_sig_sig_member(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="arrfab",
                refname="refs/heads/c7-sig-core-test",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("SIG push")

    def test_protected_sig_sig_member_precise(self):
        project = self.create_namespaced_project("rpms", "test")

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="arrfab",
                refname="refs/heads/c7-sig-core",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("SIG push")

    @patch("dist_git_auth.requests")
    def test_protected_sig_no_sig_member(self, mock_requests):
        project = self.create_namespaced_project("rpms", "test")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="pingou",
                refname="refs/heads/c7-sig-core-test",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Access to namespace rpms is restricted")

    @patch("dist_git_auth.requests")
    def test_protected_sig_sig_member_no_sig_branch(self, mock_requests):
        project = self.create_namespaced_project("rpms", "test")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="arrfab",
                refname="refs/heads/c7",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Access to namespace rpms is restricted")


class DistGitAuthTestsFedoraBranchOverride(DistGitAuthTests):
    dga_config = {
        "PR_ONLY": False,
        "ACL_BLOCK_UNSPECIFIED": False,
        "BLACKLIST_RES": ["refs/heads/c[0-9]+.*"],
        "UNSPECIFIED_BLACKLIST_RES": ["refs/heads/f[0-9]+"],
        "RCM_GROUP": "relenggroup",
        "RCM_BRANCHES": ["refs/heads/f[0-9]+"],
        "ACL_PROTECTED_NAMESPACES": ["rpms", "modules", "container"],
        "PDC_URL": "invalid://",
        "PDC_BRANCH_OVERRIDES": {"rpms": {"refs/heads/main": "refs/heads/rawhide"}},
    }

    def setUp(self):
        super(DistGitAuthTestsFedoraBranchOverride, self).setUp()

        project = self.create_namespaced_project("rpms", "test")
        project = self.create_namespaced_project(
            "container", "cockpit", hook_token_postfix="b"
        )

        msg = add_user_to_project(
            session=self.session,
            project=project,
            new_user="foo",
            user="pingou",
            access="collaborator",
            branches="epel*",
        )
        self.session.commit()
        self.assertEqual(msg, "User added")

    @patch("dist_git_auth.requests")
    def test_pushing_to_rpms_main_calls_pdc_correctly(self, mock_requests):
        project = get_project(self.session, name="test", namespace="rpms")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/main",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer: False")
        self.expect_info_msg("Fall-through deny")

        mock_requests.get.assert_called_with(
            "invalid://component-branches/?global_component=test"
            "&name=rawhide&type=rpm&fields=active"
        )

    @patch("dist_git_auth.requests")
    def test_pushing_to_rpms_rawhide_calls_pdc_correctly(self, mock_requests):
        project = get_project(self.session, name="test", namespace="rpms")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertFalse(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/rawhide",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer: False")
        self.expect_info_msg("Fall-through deny")

        mock_requests.get.assert_called_with(
            "invalid://component-branches/?global_component=test"
            "&name=rawhide&type=rpm&fields=active"
        )

    @patch("dist_git_auth.requests")
    def test_pushing_to_containers_calls_pdc_correctly(self, mock_requests):
        project = get_project(self.session, name="cockpit", namespace="container")
        res = Mock()
        res.ok = True
        res.json.return_value = {"results": []}
        mock_requests.get.return_value = res

        self.assertTrue(
            self.dga.check_acl(
                self.session,
                project=project,
                username="foo",
                refname="refs/heads/epel8",
                pull_request=None,
                repodir=None,
                repotype="main",
                revfrom=None,
                revto=None,
                is_internal=False,
            )
        )

        self.expect_info_msg("Committer: True")
        self.expect_info_msg("Committer push")

        mock_requests.get.assert_called_with(
            "invalid://component-branches/?global_component=cockpit"
            "&name=epel8&type=container&fields=active"
        )
