import json
import os
from unittest.mock import MagicMock, patch

import pagure.lib.model
import pagure.lib.query

from pagure_distgit import model
from pagure_distgit import plugin

import pygit2

import tests


class PagureFlaskApiOrphanEndpointTests(tests.Modeltests):
    """Tests the orphan endpoint added in pagure-dist-git."""

    def setUp(self):
        """Set up the environnment, ran before every tests."""
        super(PagureFlaskApiOrphanEndpointTests, self).setUp()
        self.session.flush()
        tests.create_projects(self.session)
        tests.create_projects_git(os.path.join(self.path, "repos"), bare=True)
        tests.create_tokens(self.session)
        tests.create_tokens_acl(self.session)
        tests.create_user(self.session, "orphan", "orphan", ["orphan@fedoraproject.org"])
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test3",
            namespace="somenamespace",
        )
        token = pagure.lib.query.get_api_token(self.session, "aaabbbcccddd")
        token.project = repo
        self.session.add(token)
        self.session.commit()
        self._app.register_blueprint(plugin.DISTGIT_NS)

    def test_token_missing_ACL(self):
        """Test the orphan endpoint with an API token missing the
        `modify_project` ACL.
        """
        headers = {"Authorization": "token foo_token"}
        output = self.app.post("/_dg/orphan/somenamespace/test3", headers=headers)
        # invalid token
        assert output.status_code == 401

    def test_invalid_token(self):
        """Test the orphan endpoint with an invalid API token."""
        headers = {"Authorization": "token BBBZZZOOO"}
        datainput = {}
        output = self.app.post(
            "/_dg/orphan/somenamespace/test3",
            data=datainput,
            headers=headers,
        )
        assert output.status_code == 401

    def test_empty_form(self):
        """Assert that empty form is accepted."""
        headers = {"Authorization": "token aaabbbcccddd"}
        datainput = {}
        output = self.app.post(
            "/_dg/orphan/somenamespace/test3",
            data=datainput,
            headers=headers,
        )
        assert output.status_code == 200

    def test_invalid_form(self):
        """Assert that invalid form is not accepted."""
        headers = {"Authorization": "token aaabbbcccddd"}
        datainput = {"orphan_reason": "bar"}
        output = self.app.post(
            "/_dg/orphan/somenamespace/test3",
            data=datainput,
            headers=headers,
        )
        assert output.status_code == 400

    def test_already_orphaned_package(self):
        """Assert that already orphaned package can't be orphaned again."""
        headers = {"Authorization": "token aaabbbcccddd"}
        datainput = {}
        orphan_user_obj = pagure.lib.query.get_user(self.session, "orphan")
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test3",
            namespace="somenamespace",
        )
        repo.user = orphan_user_obj
        self.session.add(repo)
        self.session.commit()
        output = self.app.post(
            "/_dg/orphan/somenamespace/test3",
            data=datainput,
            headers=headers,
        )
        assert output.status_code == 401

    @patch("pagure_distgit.plugin.pagure.lib.notify.log")
    def test_orphan_package(self, mock_log):
        """Assert that package is correctly orphaned."""
        headers = {"Authorization": "token aaabbbcccddd"}
        datainput = {"orphan_reason": "Lack of time"}
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test3",
            namespace="somenamespace",
        )
        user = repo.user
        project_user = pagure.lib.model.ProjectUser(
            project_id=repo.id,
            user_id=repo.user.id,
            access="admin",
        )
        self.session.add(project_user)
        self.session.commit()
        output = self.app.post(
            "/_dg/orphan/somenamespace/test3",
            data=datainput,
            headers=headers,
        )
        assert output.status_code == 200
        data = json.loads(output.get_data(as_text=True))
        exp = {"orphan": True, "reason": "Lack of time", "reason_info": ""}
        self.assertDictEqual(data, exp)

        assert repo.user.user == "orphan"

        watch_level = pagure.lib.query.get_watch_level_on_repo(
            self.session, user, "test3"
        )
        assert watch_level == []

        assert user not in repo.users

        assert mock_log.call_count == 2

    @patch("pagure_distgit.plugin.pagure.lib.notify.log")
    def test_orphan_package_repo_users_empty(self, mock_log):
        """Assert that package is correctly orphaned."""
        headers = {"Authorization": "token aaabbbcccddd"}
        datainput = {"orphan_reason": "Lack of time"}
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test3",
            namespace="somenamespace",
        )
        user = repo.user
        output = self.app.post(
            "/_dg/orphan/somenamespace/test3",
            data=datainput,
            headers=headers,
        )
        assert output.status_code == 200
        data = json.loads(output.get_data(as_text=True))
        exp = {"orphan": True, "reason": "Lack of time", "reason_info": ""}
        self.assertDictEqual(data, exp)
        # refresh the repo object, so we have the current state
        self.session.refresh(repo)

        assert repo.user.user == "orphan"

        watch_level = pagure.lib.query.get_watch_level_on_repo(
            self.session, user, "test3"
        )
        assert watch_level == []

        assert mock_log.call_count == 1

    def test_orphan_package_repo_no_orphan_user(self):
        """Assert that missing orphan user is correctly handled."""
        headers = {"Authorization": "token aaabbbcccddd"}
        datainput = {}
        orphan_user_obj = pagure.lib.query.get_user(self.session, "orphan")
        self.session.delete(orphan_user_obj)
        self.session.commit()

        output = self.app.post(
            "/_dg/orphan/somenamespace/test3",
            data=datainput,
            headers=headers,
        )
        assert output.status_code == 400


class PagureFlaskApiOrphanGetEndpointTests(tests.Modeltests):
    """Tests the orphan get endpoint added in pagure-dist-git."""

    def setUp(self):
        """Set up the environnment, ran before every tests."""
        super(PagureFlaskApiOrphanGetEndpointTests, self).setUp()
        self.session.flush()
        tests.create_projects(self.session)
        tests.create_projects_git(os.path.join(self.path, "repos"), bare=True)
        tests.create_tokens(self.session)
        tests.create_tokens_acl(self.session)
        tests.create_user(self.session, "orphan", "orphan", ["orphan@fedoraproject.org"])
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test3",
            namespace="somenamespace",
        )
        token = pagure.lib.query.get_api_token(self.session, "aaabbbcccddd")
        token.project = repo
        self.session.add(token)
        self.session.commit()
        self._app.register_blueprint(plugin.DISTGIT_NS)

    def test_not_orphaned(self):
        """
        Assert that correct output is returned when package is not orphaned.
        """
        output = self.app.get("/_dg/orphan/somenamespace/test3")
        assert output.status_code == 200

        data = json.loads(output.get_data(as_text=True))
        exp = {"orphan": False, "reason": "", "reason_info": ""}
        self.assertDictEqual(data, exp)

    def test_orphaned(self):
        """Assert that correct output is returned when package is orphaned."""
        orphan_user_obj = pagure.lib.query.get_user(self.session, "orphan")
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test3",
            namespace="somenamespace",
        )
        repo.user = orphan_user_obj
        reason = model.PagureOrphanReason(
            project_id=repo.id,
            reason="reason",
            reason_info="reason_info",
        )
        self.session.add(reason)
        self.session.add(repo)
        self.session.commit()
        output = self.app.get("/_dg/orphan/somenamespace/test3")
        assert output.status_code == 200

        data = json.loads(output.get_data(as_text=True))
        exp = {
            "orphan": True,
            "reason": "reason",
            "reason_info": "reason_info",
        }
        self.assertDictEqual(data, exp)

    def test_orphaned_no_reason(self):
        """Assert that correct output is returned when package is orphaned."""
        orphan_user_obj = pagure.lib.query.get_user(self.session, "orphan")
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test3",
            namespace="somenamespace",
        )
        repo.user = orphan_user_obj
        self.session.add(repo)
        self.session.commit()
        output = self.app.get("/_dg/orphan/somenamespace/test3")
        assert output.status_code == 200

        data = json.loads(output.get_data(as_text=True))
        exp = {
            "orphan": True,
            "reason": "",
            "reason_info": "",
        }
        self.assertDictEqual(data, exp)


class PagureFlaskApiTakeOrphanEndpointTests(tests.Modeltests):
    """Tests the take orphan endpoint added in pagure-dist-git."""

    def setUp(self):
        """Set up the environnment, ran before every tests."""
        super(PagureFlaskApiTakeOrphanEndpointTests, self).setUp()
        self.session.flush()
        tests.create_projects(self.session)
        tests.create_projects_git(os.path.join(self.path, "repos"), bare=True)
        tests.create_tokens(self.session)
        tests.create_tokens_acl(self.session)
        tests.create_user(self.session, "orphan", "orphan", ["orphan@fedoraproject.org"])
        pagure.lib.query.add_group(
            self.session,
            "packager",
            "packager",
            "",
            "user",
            "pingou",
            True,
            [],
        )
        pagure.lib.query.add_group(
            self.session,
            "rel-eng",
            "rel-eng",
            "",
            "user",
            "pingou",
            True,
            [],
        )

        # Create a rpms/test4 project
        item = pagure.lib.model.Project(
            user_id=1,  # pingou
            name="test4",
            namespace="rpms",
            is_fork=False,
            parent_id=None,
            description="test project #4",
            hook_token="aaabbbcccdd",
        )
        item.close_status = [
            "Invalid",
            "Insufficient data",
            "Fixed",
            "Duplicate",
        ]
        self.session.add(item)
        self.session.flush()
        tests.create_locks(self.session, item)
        self.session.commit()
        # Create its git repo
        repo_path = os.path.join(self.path, "repos", "rpms", "test4.git")
        pygit2.init_repository(repo_path, bare=True)

        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test4",
            namespace="rpms",
        )
        token = pagure.lib.query.get_api_token(self.session, "aaabbbcccddd")
        token.project = repo
        orphan_user_obj = pagure.lib.query.get_user(self.session, "orphan")
        repo.user = orphan_user_obj
        reason = model.PagureOrphanReason(
            project_id=repo.id,
            reason="reason",
            reason_info="reason_info",
        )
        self.session.add(reason)
        self.session.add(repo)
        self.session.add(token)
        self.session.commit()
        self._app.register_blueprint(plugin.DISTGIT_NS)

    def test_token_missing_ACL(self):
        """
        Test the take orphan endpoint with an API token missing
        the `modify_project` ACL.
        """
        headers = {"Authorization": "token foo_token"}
        output = self.app.post("/_dg/take_orphan/rpms/test4", headers=headers)
        # invalid token
        assert output.status_code == 401

    def test_invalid_token(self):
        """Test the take orphan endpoint with an invalid API token."""
        headers = {"Authorization": "token BBBZZZOOO"}
        output = self.app.post(
            "/_dg/take_orphan/somenamespace/test3",
            headers=headers,
        )
        assert output.status_code == 401

    @patch("pagure_distgit.plugin._is_active_in_pdc")
    @patch("pagure_distgit.plugin.pagure.lib.notify.log")
    def test_take_orphan(self, mock_log, mock_pdc):
        """Assert that package is correctly adopted."""
        mock_pdc.return_value = True
        headers = {"Authorization": "token aaabbbcccddd"}
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test4",
            namespace="rpms",
        )
        assert repo.orphan_reason.reason == "reason"
        output = self.app.post(
            "/_dg/take_orphan/rpms/test4",
            headers=headers,
        )
        assert output.status_code == 200
        data = json.loads(output.get_data(as_text=True))
        self.assertDictEqual(data, {"point_of_contact": "pingou"})

        # refresh the repo object, so we have the current state
        self.session.refresh(repo)
        assert repo.user.user == "pingou"
        assert not repo.orphan_reason

        assert mock_log.call_count == 1

    @patch.dict("pagure.config.config", {"PDC_URL": "invalid://"})
    @patch("pagure_distgit.plugin.requests")
    @patch("pagure_distgit.plugin.pagure.lib.notify.log")
    def test_take_orphan_no_reason(self, mock_log, mock_req):
        """
        Assert that package is correctly adopted when reason
        is not provided.
        """
        resp = MagicMock()
        resp.url = "http://localhost"
        resp.json.return_value = {"results": [{"active": True}]}
        mock_req.get.return_value = resp
        headers = {"Authorization": "token aaabbbcccddd"}
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test4",
            namespace="rpms",
        )
        reason = repo.orphan_reason
        self.session.delete(reason)
        self.session.commit()
        output = self.app.post(
            "/_dg/take_orphan/rpms/test4",
            headers=headers,
        )
        assert output.status_code == 200
        data = json.loads(output.get_data(as_text=True))
        self.assertDictEqual(data, {"point_of_contact": "pingou"})

        # refresh the repo object, so we have the current state
        self.session.refresh(repo)
        assert repo.user.user == "pingou"

        assert mock_log.call_count == 1

        assert mock_req.get.call_count == 1
        mock_req.get.assert_called_with(
            "invalid:/component-branches/?"
            "global_component=test4&name=rawhide&type=rpm",
            timeout=(30, 30),
        )

    @patch("pagure_distgit.plugin._is_active_in_pdc")
    def test_give_orphan_no_user(self, mock_pdc):
        """Assert that the point_of_contact cannot be updated without specifying user."""
        mock_pdc.return_value = True
        headers = {"Authorization": "token aaabbbcccddd"}
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test4",
            namespace="rpms",
        )
        output = self.app.post(
            "/_dg/give_orphan/rpms/test4",
            headers=headers,
        )
        assert output.status_code == 404
        data = json.loads(output.get_data(as_text=True))
        assert data["error"] == "No such user found"
        assert data["error_code"] == "ENOUSER"
        assert data["errors"] == "You have to specify a user as point_of_contact."

    @patch("pagure_distgit.plugin._is_active_in_pdc")
    @patch("pagure_distgit.plugin.pagure.lib.notify.log")
    def test_give_orphan(self, mock_log, mock_pdc):
        """Assert that point of contact is correctly updated."""
        mock_pdc.return_value = True
        headers = {"Authorization": "token aaabbbcccddd"}
        orphan_user_obj = pagure.lib.query.get_user(self.session, "orphan")
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test4",
            namespace="rpms",
        )

        output = self.app.post(
            "/_dg/give_orphan/rpms/test4?user=pingou",
            headers=headers,
        )
        assert output.status_code == 200
        data = json.loads(output.get_data(as_text=True))
        self.assertDictEqual(data, {"point_of_contact": "pingou"})

    @patch("pagure_distgit.plugin._is_active_in_pdc")
    @patch("pagure_distgit.plugin.pagure.lib.notify.log")
    def test_give_orphan_user_not_in_packager_group(self, mock_log, mock_pdc):
        """Assert that point of contact is correctly updated."""
        mock_pdc.return_value = True
        headers = {"Authorization": "token aaabbbcccddd"}
        orphan_user_obj = pagure.lib.query.get_user(self.session, "orphan")
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test4",
            namespace="rpms",
        )

        repo.user = orphan_user_obj
        self.session.add(repo)
        self.session.commit()

        output = self.app.post(
            "/_dg/give_orphan/rpms/test4?user=orphan",
            headers=headers,
        )
        data = json.loads(output.get_data(as_text=True))
        assert (
            data["error"]
            == "You do not have sufficient permissions to perform this action"
        )
        assert data["error_code"] == "ENOTHIGHENOUGH"
        assert data["errors"] == "The new POC must be a packager to adopt a package."

    @patch("pagure_distgit.plugin._is_active_in_pdc")
    @patch("pagure_distgit.plugin.pagure.lib.notify.log")
    def test_give_orphan_repo_user_not_orphan(self, mock_log, mock_pdc):
        """Assert that point of contact is correctly updated."""
        mock_pdc.return_value = True
        headers = {"Authorization": "token aaabbbcccddd"}
        orphan_user_obj = pagure.lib.query.get_user(self.session, "pingou")
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test4",
            namespace="rpms",
        )

        repo.user = orphan_user_obj
        self.session.add(repo)
        self.session.commit()

        output = self.app.post(
            "/_dg/give_orphan/rpms/test4?user=orphan",
            headers=headers,
        )
        data = json.loads(output.get_data(as_text=True))
        assert data["error"] == "You are not allowed to modify this project"
        assert data["error_code"] == "EMODIFYPROJECTNOTALLOWED"

    @patch("pagure_distgit.plugin._is_active_in_pdc")
    @patch("pagure_distgit.plugin.pagure.lib.notify.log")
    def test_give_orphan_user_not_in_releng_or_admin_group(self, mock_log, mock_pdc):
        """Assert that point of contact is correctly updated."""
        mock_pdc.return_value = True
        headers = {"Authorization": "token aaabbbcccddd"}
        repo = pagure.lib.query.get_authorized_project(
            self.session,
            "test4",
            namespace="rpms",
        )
        pagure.lib.query.delete_user_of_group(
           self.session,
           "pingou",
           "rel-eng",
           "pingou",
           False,
           force=True,
        )
        self.session.commit()

        output = self.app.post(
            "/_dg/give_orphan/rpms/test4?user=foo",
            headers=headers,
        )

        data = json.loads(output.get_data(as_text=True))
        assert output.status_code == 403
        assert data["errors"] == "You must be in rel-eng or admin group to assign a package."
        assert data["error_code"] == "ENOTHIGHENOUGH"
