#!/usr/bin/env python

"""
This script loads in pagure the ACLs exported from pkgdb to ensure help the
transition.


Export the pkgdb ACL database using:
````
COPY (
    SELECT DISTINCT package.namespace,
        package.name,
        package_listing_acl.fas_name,
        package_listing_acl.acl,
        package_listing_acl.status
    FROM package, package_listing, package_listing_acl, collection
    WHERE package_listing_acl.packagelisting_id = package_listing.id
    AND package_listing.package_id = package.id
    AND package_listing.collection_id = collection.id
    AND collection.status IN ('Active', 'Under Development')
    ORDER BY package.namespace, package.name,
        package_listing_acl.fas_name, package_listing_acl.acl
) TO '/tmp/pkgdb_acls.csv' delimiter ',' CSV header;
````

"""
from __future__ import print_function

import time
import os

import mock

if 'PAGURE_CONFIG' not in os.environ \
        and os.path.exists('/etc/pagure/pagure.cfg'):
    print('Using configuration file `/etc/pagure/pagure.cfg`')
    os.environ['PAGURE_CONFIG'] = '/etc/pagure/pagure.cfg'

import pagure
import pagure.lib
import pagure.lib.model


ACL_FILE = '/tmp/pkgdb_acls.csv'

mock.patch(
    'pagure.lib.notify.fedmsg_publish', mock.MagicMock(return_value=True))
mock.patch(
    'pagure.lib.update_read_only_mode', mock.MagicMock(return_value=True))
def main():
    with open(ACL_FILE) as stream:
        data = [row.strip().split(',') for row in stream]

    for row in data:
        print(row)
        namespace, package, packager, acl, status = row

        # Drop all the commit access, these should be already good
        if acl == 'commit':
            continue

        # Check if the package exists
        pkg = pagure.get_authorized_project(
            pagure.SESSION, namespace=namespace, project_name=package)
        if not pkg:
            print('ERROR: project: %s/%s not found' % (namespace, package))
            continue

        # Check if the packager exists
        user = pagure.lib.search_user(pagure.SESSION, username=packager)
        if not user:
            print('ERROR: packager: %s not found' % packager)
            continue

        if acl == 'approveacls' and status == 'Approved':
            # Check if the packager has still commit
            committers = [user.username for user in pkg.committers]
            if packager not in committers:
                print(
                    '    %s is no longer a committer on %s, let\'s not make '
                    'them admin' % (packager, pkg.fullname))
                continue
            try:
                print('    Making %s admin on %s' % (packager, pkg.fullname))
                pagure.lib.add_user_to_project(
                    session=pagure.SESSION,
                    project=pkg,
                    new_user=packager,
                    user='releng',
                    access='admin',
                    required_groups=None)
                pagure.lib.update_watch_status(
                    session=pagure.SESSION,
                    project=pkg,
                    user=packager,
                    watch='0')
            except pagure.exceptions.PagureException as err:
                print('ERROR: error while making the user %s an admin' % packager)
                print(err)
                continue
        elif acl.startswith('watch'):
            watch_levels = pagure.lib.get_watch_level_on_repo(
                pagure.SESSION,
                packager,
                package,
                namespace=namespace)
            print('   ', acl, status, watch_levels)

            """watch status can be:
            -1: reset the watch status to default
            0: unwatch, don't notify the user of anything
            1: watch issues and PRs
            2: watch commits
            3: watch issues, PRs and commits
            """
            watch = None
            if acl == 'watchbugzilla':
                if 'issues' in watch_levels \
                        and status not in ['Approved', 'Awaiting Review']:
                    # Remove watch bugzilla
                    print('    Removing watch bugzilla/issues')
                    watch = '0'
                    if 'commits' in watch_levels:
                        watch = '2'
                elif 'issues' not in watch_levels \
                        and status == 'Approved':
                    # Add watch bugzilla
                    print('    Adding watch bugzilla')
                    watch = '1'
                    if 'commits' in watch_levels:
                        watch = '3'
                else:
                    print('    Nothing to do for watch bugzilla')

            elif acl == 'watchcommits':
                if 'commits' in watch_levels \
                        and status not in ['Approved', 'Awaiting Review']:
                    # Remove watch commits
                    print('    Removing watch commits')
                    watch = '0'
                    if 'issues' in watch_levels:
                        watch = '1'
                elif 'commits' not in watch_levels \
                        and status == 'Approved':
                    # Add watch commits
                    print('    Adding watch commits')
                    watch = '2'
                    if 'issues' in watch_levels:
                        watch = '3'
                else:
                    print('    Nothing to do for watch commits')

            if watch:
                print(
                    'Setting watch level: %s to %s for %s' % (
                        watch, packager, pkg.fullname))
                pagure.lib.update_watch_status(
                    session=pagure.SESSION,
                    project=pkg,
                    user=packager,
                    watch=watch)

        pagure.SESSION.commit()


if __name__ == '__main__':
    import sys
    sys.exit(main())
